/* FILE: oxsthread.h              -*-Mode: c++-*-
 *
 * OXS thread support.
 *
 * These classes are only compiled if the OXS_DO_THREADS macro
 * is defined.
 */

#ifndef _OXS_THREAD
#define _OXS_THREAD

#if OOMMF_THREADS

#include <tcl.h>
#include <vector>

#include "oc.h"

/* End includes */

// Thread-safe printf
void Oxs_ThreadPrintf(FILE* fptr,const char *format, ...);

// Base class that Oxs_Thread client should inherit from.
class Oxs_ThreadRunObj {
public:
  virtual void Cmd(int threadnumber,void* data) =0;
  virtual ~Oxs_ThreadRunObj() {}
};

// Thread control trio: mutex, condition, counter
struct Oxs_ThreadControl {
  Tcl_Mutex mutex;      // In general, this mutex should be acquired
  Tcl_Condition cond;   // changing either cond or count.
  int count;            // count is used to protect against spurious
  /// notifies; if count>0 then a thread is waiting on cond.
  void Finalize() {
    Tcl_ConditionFinalize(&cond);  cond=0;
    Tcl_MutexFinalize(&mutex);     mutex=0;
    count=0;
  }
  Oxs_ThreadControl() : mutex(0), cond(0), count(0) {}
  ~Oxs_ThreadControl() { Finalize(); }
};

// Object for representation/control of actual thread
class Oxs_Thread {
 public:
  Oxs_Thread(int thread_number);
  ~Oxs_Thread();
  void RunCmd(Oxs_ThreadControl& stop,Oxs_ThreadRunObj& runobj,void* data);

  // When start.count>0, control OS thread is paused, waiting for a
  // new command to run.  start.count==0 means thread is active.
  int IsRunning() { return (0==start.count); }

 private:
  friend Tcl_ThreadCreateProc _Oxs_Thread_threadmain;
  Tcl_ThreadId thread_id;   // Assigned by Tcl and/or OS
  int thread_number;        // Assigned by Oxs_ThreadTree static control
  Oxs_ThreadControl start;  // Unique to thread object
  Oxs_ThreadControl* stop;  // Assigned by and unique to tree object
  Oxs_ThreadRunObj* runobj; // Assigned by user code through Oxs_ThreadTree
  void* data;               // Ditto

  // Disable copy constructor and operator=() by declaring
  // without providing implementations.
  Oxs_Thread(const Oxs_Thread&);
  Oxs_Thread& operator=(const Oxs_Thread&);

};

// Manager for Oxs_Thread objects.  Call Oxs_ThreadTree to
// start and stop threads.
class Oxs_ThreadTree {
public:
  Oxs_ThreadTree() : threads_unjoined(0) {}
  ~Oxs_ThreadTree() {} // The mutex and condition variables in stop
  /// are automatically finalized inside the Oxs_ThreadControl
  /// destructor.

  static void EndThreads();

  void Launch(Oxs_ThreadRunObj& runobj,void* data); // Launch
  /// threads, one at a time.
  void Join(); // Joins all running threads.

private:
  Oxs_ThreadControl stop; // One per tree
  int threads_unjoined;   // Count of launched, unjoined threads.
  /// Mirrors stop.count, except modified only by *this, not
  /// client threads.

  static Tcl_Mutex launch_mutex;
  static std::vector<Oxs_Thread*> threads;

  // Disable copy constructor and operator=() by declaring
  // without providing implementations.
  Oxs_ThreadTree(const Oxs_ThreadTree&);
  Oxs_ThreadTree& operator=(const Oxs_ThreadTree&);
};


#endif // OOMMF_THREADS
#endif // _OXS_THREAD
